import numpy as np

def compute_intervals(coords, coord_labels, max_time_between=86400, distance_metric="haversine"):
    """Compute stop and moves intervals from the list of labels.
    
    Parameters
    ----------
        coords : np.array (shape=(N, 2) or shape=(N,3))
        coord_labels: 1d np.array of integers

    Returns
    -------
        intervals : array-like (shape=(N_intervals, 5))
            Columns are "label", "start_time", "end_time", "latitude", "longitude"
    
    """
    
    if coords.shape[1] == 2:
        times = np.array(list(range(0, len(coords))))
        coords = np.hstack([coords, times.reshape(-1,1)])
        
    trajectory = np.hstack([coords, coord_labels.reshape(-1,1)])
    
    final_trajectory = []
    
    #initialize values
    lat_prec, lon_prec, t_start, loc_prec = trajectory[0]  
    t_end = t_start 
    median_lat = [lat_prec]
    median_lon = [lon_prec]

    #Loop through trajectory
    for lat, lon, time, loc in trajectory[1:]:
        
        #if the location name has not changed update the end of the interval
        if (loc==loc_prec) and (time-t_end)<(max_time_between):
            t_end = time
            median_lat.append(lat)
            median_lon.append(lon)
            
            
        #if the location name has changed build the interval and reset values
        else:
            if loc_prec==-1:
                final_trajectory.append([loc_prec, t_start,  t_end, np.nan, np.nan])
            else:
                final_trajectory.append([loc_prec, t_start,  t_end, np.median(median_lat), np.median(median_lon)])
                
            t_start = time 
            t_end = time 
            median_lat = []
            median_lon = []
            
        
        #update current values
        loc_prec = loc
        lat_prec = lat
        lon_prec = lon
        
    #Add last group
    if loc_prec==-1:
        final_trajectory.append([loc_prec, t_start,  t_end, np.nan, np.nan])
    else:
        final_trajectory.append([loc_prec, t_start,  t_end, np.median(median_lat), np.median(median_lon)])

    return final_trajectory


def label_trace(*args):
	raise DeprecationWarning("""`label_trace` is no longer supported. Instead use:
			>>> from infostop import Infostop
			>>> model = Infostop()
			>>> labels = model.fit_predict(my_data)""")
	
def label_static_points(*args):
	raise DeprecationWarning("""`label_static_points` is no longer supported. Instead use:
			>>> from infostop import SpatialInfomap
			>>> model = SpatialInfomap()
			>>> labels = model.fit_predict(my_data)""")

def label_network(*args):
	raise DeprecationWarning("""`label_network` is no longer supported. Instead use:
			>>> from infostop import Infostop
			>>> model = Infostop()
			>>> labels = model.fit_predict(my_data)""")


#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This package implements stop location detection using Infomap.
"""

from .deprecated import label_trace
from .deprecated import label_static_points
from .deprecated import label_network
from .models import Infostop
from .models import SpatialInfomap
from .visualize import plot_map
from .postprocess import compute_intervals
import cpputils

from .metadata import __version__, __name__



import numpy as np
from sklearn.neighbors import BallTree
from scipy.spatial.qhull import QhullError
from infomap import Infomap
from scipy.spatial import ConvexHull
from tqdm import tqdm

def pass_func(input, **kwargs):
    return input

def query_neighbors(coords, r2, distance_metric='haversine', weighted=False):
    """Build a network from a set of points and a threshold distance.
    
    Parameters
    ----------
        coords : array-like (N, 2)
        r2 : float
            Threshold distance.
        distance_metric : str
            Either 'haversine' or None.
    
    Returns
    -------
        nodes : list of ints
            Correspond to the list of nodes
        edges : list of tuples
            An edge between two nodes exist if they are closer than r2.
        singleton nodes : list of ints
            Nodes that have no connections, e.g. have been visited once.
    """
    
    # If the metric is haversine update points (to radians) and r2 accordingly.
    if distance_metric == 'haversine':
        coords = np.radians(coords)
        r2 = r2 / 6371000

    # Init tree
    tree = BallTree(coords, metric=distance_metric)

    # Query
    return tree.query_radius(coords, r=r2, return_distance=weighted)


def infomap_communities(node_idx_neighbors, node_idx_distances, counts, weight_exponent, distance_metric, verbose):
    """Two-level partition of single-layer network with Infomap.
    
    Parameters
    ----------
        node_index_neighbors : array of arrays
            Example: `array([array([0]), array([1]), array([2]), ..., array([9997]),
       array([9998]), array([9999])], dtype=object)`.

    Returns
    -------
        out : dict (node-community hash map).
    """
    # Tracking 
    if verbose: progress = tqdm
    else:       progress = pass_func

    # Initiate  two-level Infomap
    network = Infomap("--two-level")

    # Add nodes (and reindex nodes because Infomap wants ranked indices)
    if verbose: print("    ... adding nodes:")
    name_map, name_map_inverse = {}, {}
    singleton_nodes = []
    infomap_idx = 0
    for n, neighbors in progress(enumerate(node_idx_neighbors), total=len(node_idx_neighbors)):
        if len(neighbors) > 1:
            network.addNode(infomap_idx)
            name_map_inverse[infomap_idx] = n
            name_map[n] = infomap_idx
            infomap_idx += 1
        else:
            singleton_nodes.append(n)

    if verbose:
        print(f"    --> added {len(name_map)} nodes (found {len(singleton_nodes)} singleton nodes)")

    
    # Add links
    if verbose:
        n_edges = 0
        print("    ... adding edges")

    if node_idx_distances is None:
        for node, neighbors in progress(enumerate(node_idx_neighbors), total=len(node_idx_neighbors)):
            for neighbor in neighbors[neighbors > node]:
                network.addLink(name_map[node], name_map[neighbor], max(counts[node], counts[neighbor]))
                if verbose: n_edges += 1
    else:
        for node, (neighbors, distances) in progress(enumerate(zip(node_idx_neighbors, node_idx_distances)), total=len(node_idx_neighbors)):
            for neighbor, distance in zip(neighbors[neighbors > node], distances[neighbors > node]):
                if distance_metric == "haversine":
                    distance *= 6371000
                network.addLink(name_map[node], name_map[neighbor], max(counts[node], counts[neighbor]) * distance**(-weight_exponent))
                if verbose: n_edges += 1
    
    if verbose:
        print(f"    --> added {n_edges} edges")

    # Run infomap
    if verbose: print("    ... running Infomap...", end=" ")
    if len(name_map) > 0:
        network.run()
    	# Convert to node-community dict format
        partition = dict([
            (name_map_inverse[infomap_idx], module)
            for infomap_idx, module in network.modules
        ])
        if verbose: print("done")
    else:
        partition = {}


    if verbose:
        print(f"Found {len(set(partition.values()))-1} stop locations")

    return partition, singleton_nodes


def label_network(node_idx_neighbors, node_idx_distances, counts, weight_exponent, label_singleton, distance_metric, verbose):
    """Infer infomap clusters from distance matrix and link distance threshold.
    
    Parameters
    ----------
        nodes: array 
            Nodes in the network.
        edges: array
            Edges in the network (two nodes are connected if distance<r2).
        singleton_nodes: array
            Non connected nodes.
        label_singleton: bool
            If True, give stationary locations that was only visited once their own
            label. If False, label them as outliers (-1).
            
    Returns
    -------
        out : array-like (N, )
            Array of labels matching input in length. Detected stop locations are labeled from 0
            and up, and typically locations with more observations have lower indices. If
            `label_singleton=False`, coordinates with no neighbors within distance `r2` are
            labeled -1.
    """ 
    # Infer the partition with infomap. Partiton looks like `{node: community, ...}`
    partition, singleton_nodes = infomap_communities(node_idx_neighbors, node_idx_distances, counts, weight_exponent, distance_metric, verbose)
    
    # Add new labels to each singleton point (stop that was further than r2 from
    # any other point and thus was not represented in the network)
    if label_singleton:
        max_label = max(partition.values(), default=-1)
        partition.update(dict(zip(
            singleton_nodes,
            range(max_label+1, max_label+1+len(singleton_nodes))
        )))
    
    # Cast the partition as a vector of labels like `[0, 1, 0, 3, 0, 0, 2, ...]`
    return np.array([
        partition[n] if n in partition else -1
        for n in range(len(node_idx_neighbors))
    ])

def max_pdist(points):
    """ 
    Calculate the distance bewteen each pair in a set of points given a distance function.
    
    Author: Piotr Sapiezynski
    Source: https://github.com/sapiezynski/haversinevec

    Input
    -----
        points : array-like (shape=(N, 2))
            (lat, lon) in degree or radians (default is degree)

    Output
    ------
        result : array-like (shape=(N*(N-1)//2, ))
    """ 
    def _l2(points_a, points_b):
        return np.linalg.norm((points_a - points_b).reshape(-1,2),axis = 1)

    c = points.shape[0]
    result = np.zeros((c*(c-1)//2,), dtype=np.float64)
    vec_idx = 0
    for idx in range(0, c-1):
        ref = points[idx]
        temp = _l2(points[idx+1:c, :], ref)
        #to be taken care of
        result[vec_idx:vec_idx+temp.shape[0]] = temp
        vec_idx += temp.shape[0]
    return max(result)

def convex_hull(points, to_return='points'):
    """Return the convex hull of a collection of points."""
    try:
        hull = ConvexHull(points)
        return points[hull.vertices, :]
    except QhullError:
        c = points.mean(0)
        if points.shape[0] == 1:
            l = 5e-5
        else:
            l = max_pdist(points)
        return np.vstack([
            c + np.array([-l/2, -l/2]),  # bottom left
            c + np.array([l/2, -l/2]),   # bottom right
            c + np.array([l/2, l/2]),    # top right
            c + np.array([-l/2, l/2]),    # top right
        ])



import folium
from folium.plugins import HeatMap
import numpy as np
from infostop import utils

class FoliumMap:
    """Plot points on a folium map.

    Parameters
    ----------
        points : 2d numpy.array (shape=(N, 2))
        labels : 1d numpy.array (shape=(N, ))
        zoom_start : int/float
            Initial zoom of the map.
        tiles : str 
            Which background tiles to use in the map. Any valid tiles for `folium.Map` method.
            Default is 'OpenStreetMap'. A good minimalist alternative is 'CartoDBpositron'.
        API_key : str
            Mapbox API key.
    
    Example
    -------
        >>> folmap = FoliumMap(mypoints, mylabels)
        >>> folmap.render_polygons()
        >>> folmap.render_heatmap()
        >>> folmap.m
    """
    def __init__(self, points, labels, zoom_start=10, tiles='OpenStreetMap', API_key=None):
        self.points = points
        self.labels = labels
        self.m = folium.Map(
            location=tuple(np.median(points, axis=0)),
            zoom_start=zoom_start,
            tiles=tiles,
            API_key=API_key,
            attr="Infostop"
        )


    def render_polygons(self, color="#ee9999", opacity=0.3):
        """Render convex hulls of points in each label-group.

        Parameters
        ----------
            color : str 
                Color of stop location areas.
            opacity : float (in [0, 1])
                Opacity of stop location areas.
        """
        def _style_function(feature):
            return {"fillColor": feature["properties"]["color"], "color": feature["properties"]["color"], "weight": 1, "fillOpacity": opacity}

        stop_hulls = []
        for stop_idx in (set(self.labels) - {-1}):
            p = self.points[self.labels == stop_idx]
            stop_hulls.append(utils.convex_hull(p))

        features = {"type": "FeatureCollection", "features": [
            {
                "type": "Feature",
                "properties": {"color": color},
                "geometry": {
                    "type": "Polygon",
                    "coordinates": []
                }
            }
        ]}

        for hull in stop_hulls:
            features["features"][0]["geometry"]["coordinates"].append(
                hull[:, ::-1].tolist()
            )

        folium.GeoJson(
            features,
            style_function=_style_function
        ).add_to(self.m)


    def render_points(self, color="k", opacity=0.3, subsampling=1):
        """Render points.

        Parameters
        ----------
            color : str
                Color of points.
            opacity : float (in [0, 1])
                Opacity of points.
            subsampling : float (in [0, 1])
                Sampling rate of scattered points. For example, using 0.2 shows only 20 percent
                of the data.
        """
        p = self.points
        if subsampling < 1:
            mask = np.random.choice(range(p.shape[0]), size=int(subsampling * p.shape[0]), replace=False)
            p = p[mask]

        for d in p:
            folium.CircleMarker(
                d[:2],
                color=color,
                fill_color=color,
                fill_opacity=opacity,
                fill=True
            ).add_to(self.m)

    def render_heatmap(self, radius=8, subsampling=1):
        """Render points.

        Parameters
        ----------
            radius : int/float
                Kernel radius of points rendered in the heatmap. It makes sense to adjust inversely
                to the density of points, but in general [8, 10] gives good results.
            subsampling : float (in [0, 1])
                Sampling rate of points used for rendering the heatmap. For example, using 0.2
                renders a heatmap based on only 20 percent of the data.
        """
        p = self.points
        if subsampling < 1:
            mask = np.random.choice(range(p.shape[0]), size=int(subsampling * p.shape[0]), replace=False)
            p = p[mask]
        HeatMap(p, radius=radius).add_to(self.m)


def plot_map(model, display_data="unique_stationary", polygons=True, scatter=False, heatmap=True, polygons_color="#ee9999", polygons_opacity=0.3, scatter_color="k", scatter_opacity=0.3, scatter_subsampling=1, heatmap_radius=8, heatmap_subsampling=1, zoom_start=12, tiles='OpenStreetMap', API_key=None):
    """Load Folium map to display the data and fit.

    Note: `plot_map` only works if `distance_metric` is "havesine".

    Parameters
    ----------
        model : fitted instance of `Infostop` or `SpatialInfomap`
        display_data : str 
            Which data to display. Defaults to "unique_stationary" which is the downsampled data
            that Infostop is fitted on (the medians of all stationary location events, where
            duplicates have been removed). Other options are "all", which displays all of the
            original input data, "all_stationary" which displays all the input data except points
            that are labeled as non-stationary, and "all_nonstationary" which displays all the
            non-stationary labeled input data. 
        polygons : bool
            Display the convex hull polygons representing fitted stop locations. 
        scatter : bool
            Display points (slow: not recommended for over 10.000 points). 
        heatmap : bool
            Display a heatmp (typically a better option than `scatter==True`).
        polygons_color : str 
            Color of stop location areas.
        polygons_opacity : float (in [0, 1])
            Opacity of stop location areas.
        scatter_color : str
            Color of points.
        scatter_opacity : float (in [0, 1])
            Opacity of points.
        scatter_subsampling : float (in [0, 1])
            Sampling rate of scattered points. For example, using 0.2 shows only 20 percent
            of the data.
        heatmap_radius : int/float
            Kernel radius of points rendered in the heatmap. It makes sense to adjust inversely
            to the density of points, but in general [8, 10] gives good results.
        heatmap_subsampling : float (in [0, 1])
            Sampling rate of points used for rendering the heatmap. For example, using 0.2
            renders a heatmap based on only 20 percent of the data.
        zoom_start : int/float
            Initial zoom of the map.
        tiles : str 
            Which background tiles to use in the map. Any valid tiles for `folium.Map` method.
            Default is 'OpenStreetMap'. A good minimalist alternative is 'CartoDBpositron'.
        API_key : str
            Mapbox API key.

    Returns
    -------
        folmap : folium.folium.Map

    Example
    -------
        >>> # create model and fit it to data
        >>> model = Infostop()
        >>> _ = model.fit_predict(traces)
        >>> # visualize result
        >>> folmap = visualize.plot_map(model)
        >>> folmap.m
    """
    assert hasattr(model, "is_fitted"), \
           "`model` must either be an instance of `Infostop` or `SpatialInfomap`."
    assert model.is_fitted, \
           "It appears that no data is fitted to the model."
    assert model.distance_metric == "haversine", \
           "`distance_metric` is not 'haversine'."
    assert display_data in ['all', 'all_stationary', 'all_nonstationary', 'unique_stationary'], \
           'Keyword argument `display_data` must be one of "all", "all_stationary", "all_nonstationary" and "unique_stationary".'

    # If it's an instance of Infostop
    if hasattr(model, "r1"):
        if "all" in display_data:
            data_unique, indices_unique = np.unique(np.vstack(model.data)[:, :2], axis=0, return_index=True)
            labels_unique = np.hstack(model.coord_labels)[indices_unique]
            if display_data == "all":
                points = data_unique
                labels = labels_unique
            elif display_data == "all_stationary":
                points = data_unique[labels_unique>=0]
                labels = labels_unique[labels_unique>=0]
            elif display_data == "all_nonstationary":
                points = data_unique[labels_unique==-1]
                labels = labels_unique[labels_unique==-1]
        elif display_data == "unique_stationary":
            points = model.stat_coords[model.stat_labels >= 0]
            labels = model.stat_labels[model.stat_labels >= 0]

    # If it's an instance of SpatialInfomap
    else:
        points = model.stat_coords[model.stat_labels >= 0]
        labels = model.stat_labels[model.stat_labels >= 0]

    folmap = FoliumMap(
        points=points,
        labels=labels,
        zoom_start=zoom_start,
        tiles=tiles,
        API_key=API_key
    )

    if polygons:
        folmap.render_polygons(color=polygons_color, opacity=polygons_opacity)
    if scatter:
        folmap.render_points(color=scatter_color, opacity=scatter_opacity, subsampling=scatter_subsampling)
    if heatmap:
        folmap.render_heatmap(radius=heatmap_radius, subsampling=heatmap_subsampling)

    return folmap


# -*- coding: utf-8 -*-
"""
Package matadata
"""

__version__ = "0.1.6"
__name__ = "infostop"
__author__ = "Ulf Aslak, Laura Alessandretti"
__copyright__ = "Copyright 2020, Ulf Aslak"
__credits__ = ["Ulf Aslak", "Laura Alessandretti"]
__license__ = "MIT"
__maintainer__ = "Ulf Aslak"
__email__ = "ulfjensen@gmail.com"
__status__ = "Development"



import numpy as np
import cpputils
from infostop import utils
from tqdm import tqdm
 

class Infostop:
    """Infer stop-location labels from mobility trace. Dynamic points are labeled -1.

    Note: currently ONLY works for 2-dimensional spatial data. Data columns 0 and 1 are
    always intepreted as spatial locations while column 2 is interpreted as time.

    The method entils the following steps:
        1.  Sequential downsampling: Detect which points are stationary and store only
            the median of each stationarity event. A point belongs to a stationarity
            event if it is less than `r1` distance units away from the median of the
            time-previous collection of stationary points.
        2.  Spatial downsampling: Remove duplicate coordinates. Optionally round off
            coordinates before removing duplicates, to increase the downsampling.
        3.  For each coordinate, find its neighbors within `r2`.
        4.  Create Infomap network and optimize clusters in two-levels.
        5.  Reverse downsampling from step 2.
        6.  Reverse downsampling from step 1.
    
    Parameters
    ----------
        r1 : int/float
            Max distance between time-consecutive points to label them as stationary
        r2 : int/float
            Max distance between stationary points to form an edge.
        label_singleton: bool
            If True, give stationary locations that was only visited once their own label. If False, 
            label them as non-stationary (-1).
        min_staying_time : int
            The shortest duration that can constitute a stop.
        max_time_between : int
            The longest duration that can constitute a stop.
        min_size : int
            Minimum size of group to consider it stationary
        min_spacial_resolution : float
            The minimal difference allowed between points before they are considered the same points.
            Highly useful for spatially downsampling data and which dramatically reduces runtime. Higher
            values yields higher downsampling. For geo data, it is not recommended to increase it beyond
            1e-4. 1e-5, typically works well and has little impact on results.
        distance_metric : str
            Either 'haversine' (for geo data) or 'euclidean'
        weighted : bool
            Weight edges in the network representation by distance. Specifically, the weight used is
            `1 / distance(a, b) ** weight_exponent`.
        weight_exponent : float
            Exponent used when weighting edges in the network. Higher values increases the link strength
            between nearby locations yielding smaller stop locations. It is not recommended to use a negative
            `weight_exponent`, unless for a special purpose.
        verbose : bool
            Print output during the fitting procedure. Mostly for development and debugging. Option will
            be removed in future versions.

    Returns
    -------
        self : object

    Example
    -------
        >>> model = Infostop()
        >>> labels = model.fit_predict(traces)
    """
    def __init__(
        self,
        r1=10,
        r2=10,
        label_singleton=True,
        min_staying_time=300,
        max_time_between=86400,
        min_size=2,
        min_spacial_resolution=0,
        distance_metric='haversine',
        weighted=False,
        weight_exponent=1,
        verbose=False
    ):
        # Set input parameters as attributes
        self.r1 = r1
        self.r2 = r2
        self.label_singleton = label_singleton
        self.min_staying_time = min_staying_time
        self.max_time_between = max_time_between
        self.min_size = min_size
        self.min_spacial_resolution = min_spacial_resolution
        self.distance_metric = distance_metric
        self.weighted = weighted
        self.weight_exponent = weight_exponent
        self.verbose = verbose
        self.data = None
        self.stat_coords = None
        self.stat_labels = None
        self.counts = None
        self.labels = None
        self.is_fitted = False

        # Run hyper parameter assertions
        self._hyperparam_assertions()


    def fit_predict(self, data):
        """Fit Infostop on one or more location sequnces, and return labels.

        Parameters
        ----------
            data : numpy.array (shape (N, 2)/(N, 3)) or list of such numpy.arrays
                Columns 0 and 1 are reserved for lat and lon. Column 2 is reserved for time (any unit consistent with
                `min_staying_time` and `max_time_between`). If the input type is a list of arrays, each array is assumed
                to be the trace of a single user, in which case the obtained stop locations are shared by all users in
                the population.

        Returns
        -------
            coord_labels : 1d numpy.array or list of such

        Example
        -------
            >>> model = Infostop()
            >>> labels = model.fit_predict(traces)
            >>> assert type(traces) == type(labels)
        """
        
        if self.verbose: progress = tqdm
        else:            progress = utils.pass_func

        # Infer multiuser mode
        self.multiuser = True
        if type(data) != list:
            self.data = [data]
            self.multiuser = False
            progress = utils.pass_func  # no need to log progress in (1) if there's just one user
        else:
            self.data = data
            if len(data) == 1:
                progress = utils.pass_func

        if self.verbose:
            print('Multiuser input:', self.multiuser)

        # Assert the input data
        self._data_assertions(self.data)

        # (1) Sequential downsampling: group time-adjacent points
        if self.verbose:
            avg_reduction = []
            print("Downsampling in time: keeping medians of stationary events")

        stop_events, event_maps = [], []
        for u, coords_u in progress(enumerate(self.data), total=len(self.data)):
            stop_events_u, event_map_u = cpputils.get_stationary_events(
                coords_u, self.r1, self.min_size, self.min_staying_time,
                self.max_time_between, self.distance_metric
            )

            if self.verbose:
                avg_reduction.append((1 - len(stop_events_u) / len(coords_u)) * 100)

            stop_events.append(stop_events_u)
            event_maps.append(event_map_u)

        if self.verbose:
            print("    --> %sreduction was %.1f%%" % ("average " if self.multiuser else "", np.mean(avg_reduction)))
        
        # Merge `stop_events` from different users into `stat_coords`
        try:
            self.stat_coords = np.vstack([se for se in stop_events if len(se) > 0])
        except ValueError:
            raise Exception("No stop events found. Check that `r1`, `min_staying_time` and `min_size` parameters are chosen correctly.")

        # (2) Downsample (dramatically reduces computation time)
        if self.min_spacial_resolution > 0:
            self.stat_coords = np.around(self.stat_coords / self.min_spacial_resolution) * self.min_spacial_resolution

        if self.verbose:
            num_stat_orig = len(self.stat_coords)
            print(f"Downsampling {num_stat_orig} total stop events to...", end=" ")

        # Only keep unique coordinates for clustering
        self.stat_coords, inverse_indices, self.counts = np.unique(
            self.stat_coords,
            return_inverse=True, return_counts=True, axis=0
        )

        if self.verbose:
            print(f"{len(self.stat_coords)}", end=" ")
            print("(%.1f%% duplicates)" % ((1 - len(self.stat_coords)/num_stat_orig)*100))

        # (3) Find neighbors within `r2` for each point
        if self.verbose:
            print("Finding neighbors...", end=" ")
        
        ball_tree_result = utils.query_neighbors(self.stat_coords, self.r2, self.distance_metric, self.weighted)
        
        if self.weighted:
            node_idx_neighbors, node_idx_distances = ball_tree_result
        else:
            node_idx_neighbors, node_idx_distances = ball_tree_result, None

        if self.verbose:
            print("done")
            
        # (4) Create network and run infomap
        if self.verbose: print("Creating network and clustering with Infomap...")
        self.stat_labels = utils.label_network(
            node_idx_neighbors, node_idx_distances, self.counts, self.weight_exponent,
            self.label_singleton, self.distance_metric, self.verbose
        )

        # (5) Reverse the downsampling in step (2)
        self.labels = self.stat_labels[inverse_indices]
        
        # (6) Reverse the downsampling in step (1)
        self.coord_labels = []
        for j, event_map_u in enumerate(event_maps):
            i0 = sum([len(stop_events[j_]) for j_ in range(j)])
            i1 = sum([len(stop_events[j_]) for j_ in range(j+1)])
            labels_u = np.hstack([self.labels[i0:i1], -1])
            self.coord_labels.append(labels_u[event_map_u])

        # Update model state and return labels
        self.is_fitted = True
        if self.multiuser:
            return self.coord_labels
        else:
            return self.coord_labels[0]


    def _hyperparam_assertions(self):
        assert self.r1 > 0, \
               "`r1` must be > 0"
        assert self.r2 > 0, \
               "`r2` must be > 0"
        assert type(self.label_singleton) is bool, \
               "`label_singleton` either `True` or `False`"
        assert self.min_staying_time > 0, \
               "`min_staying_time` must be > 0"
        assert self.max_time_between > 0, \
               "`max_time_between` must be > 0"
        assert self.max_time_between > self.min_staying_time, \
               "`max_time_between` must be > min_staying_time"
        assert self.min_size > 1, \
               "`min_size` must be > 1"
        assert 0 <= self.min_spacial_resolution <= 1, \
               "`min_spacial_resolution` must be within [0, 1]"
        assert self.distance_metric in ['euclidean', 'haversine'], \
               "`distance_metric` should be either 'euclidean' or 'haversine'"


    def _data_assertions(self, data):
        assert not np.any(np.isnan(np.vstack(data))), \
               f"There are {np.isnan(np.vstack(data))} NaN values in the input data."
        for u, coords_u in enumerate(data):
            error_insert = "" if not self.multiuser else f"User {u}: "
            assert coords_u.shape[1] in [2, 3], \
                   "%sNumber of columns must be 2 or 3" % error_insert
            if coords_u.shape[1] == 3:
                assert np.all(coords_u[:-1, 2] <= coords_u[1:, 2]), \
                       "%sTimestamps must be ordered" % error_insert
            if self.distance_metric == 'haversine':
                assert (np.min(coords_u[:, 0]) > -90 and np.max(coords_u[:, 0]) < 90), \
                       "%sLatitude (column 0) must have values between -90 and 90" % error_insert
                assert (np.min(coords_u[:, 1]) > -180 and np.max(coords_u[:, 1]) < 180), \
                       "%sLongitude (column 1) must have values between -180 and 180" % error_insert


class SpatialInfomap:
    """Cluster a collection of points using Infomap.

    The method entils the following steps:
        1.  Spatial downsampling: Remove duplicate coordinates. Optionally round off
            coordinates before removing duplicates, to increase the downsampling.
        2.  For each coordinate, find its neighbors within `r2`.
        3.  Create Infomap network and optimize clusters in two-levels.
        4.  Reverse downsampling from step 1.
    
    Parameters
    ----------
        r2 : int/float
            Max distance between stationary points to form an edge.
        label_singleton: bool
            If True, give stationary locations that was only visited once their own label. If False, 
            label them as non-stationary (-1).
        min_spacial_resolution : float
            The minimal difference allowed between points before they are considered the same points.
            Highly useful for spatially downsampling data and which dramatically reduces runtime. Higher
            values yields higher downsampling. For geo data, it is not recommended to increase it beyond
            1e-4. 1e-5, typically works well and has little impact on results.
        distance_metric : str
            Either 'haversine' (for geo data) or 'euclidean'
        weighted : bool
            Weight edges in the network representation by distance. Specifically, the weight used is
            `1 / distance(a, b) ** weight_exponent`.
        weight_exponent : float
            Exponent used when weighting edges in the network. Higher values increases the link strength
            between nearby locations yielding smaller stop locations. It is not recommended to use a negative
            `weight_exponent`, unless for a special purpose.
        verbose : bool
            Print output during the fitting procedure. Mostly for development and debugging. Option will
            be removed in future versions.

    Returns
    -------
        self : object

    Example
    -------
        >>> model = SpacialInfomap()
        >>> labels = model.fit_predict(coordinates)
    """
    def __init__(
        self,
        r2=10,
        label_singleton=True,
        min_spacial_resolution=0,
        distance_metric='haversine',
        weighted=False,
        weight_exponent=1,
        verbose=False
    ):
        # Set input parameters as attributes
        self.r2 = r2
        self.label_singleton = label_singleton
        self.min_spacial_resolution = min_spacial_resolution
        self.distance_metric = distance_metric
        self.weighted = weighted
        self.weight_exponent = weight_exponent
        self.verbose = verbose
        self.data = None
        self.stat_coords = None
        self.stat_labels = None
        self.counts = None
        self.labels = None
        self.is_fitted = False

        # Run hyper parameter assertions
        self._hyperparam_assertions()


    def fit_predict(self, data):
        """Fit Infostop on one or more location sequnces, and return labels.

        Parameters
        ----------
            data : numpy.array (shape (N, 2)/(N, 3)) or list of such numpy.arrays
                Columns 0 and 1 are reserved for lat and lon. Column 2 is reserved for time (any unit consistent with
                `min_staying_time` and `max_time_between`). If the input type is a list of arrays, each array is assumed
                to be the trace of a single user, in which case the obtained stop locations are shared by all users in
                the population.

        Returns
        -------
            coord_labels : 1d numpy.array or list of such

        Example
        -------
            >>> model = Infostop()
            >>> labels = model.fit_predict(traces)
            >>> assert type(traces) == type(labels)
        """
        
        if self.verbose: progress = tqdm
        else:            progress = utils.pass_func

        self.data = data

        # Assert the input data
        self._data_assertions(self.data)
        
        # (1) Downsample (dramatically reduces computation time)
        self.stat_coords = data
        if self.min_spacial_resolution > 0:
            self.stat_coords = np.around(self.stat_coords / self.min_spacial_resolution) * self.min_spacial_resolution

        if self.verbose:
            num_stat_orig = len(self.stat_coords)
            print(f"Downsampling {num_stat_orig} total stop events to...", end=" ")

        # Only keep unique coordinates for clustering
        self.stat_coords, inverse_indices, self.counts = np.unique(
            self.stat_coords,
            return_inverse=True, return_counts=True, axis=0
        )

        if self.verbose:
            print(f"{len(self.stat_coords)}", end=" ")
            print("(%.1f%% duplicates)" % ((1 - len(self.stat_coords)/num_stat_orig)*100))

        # (2) Find neighbors within `r2` for each point
        if self.verbose:
            print("Finding neighbors...", end=" ")
        
        ball_tree_result = utils.query_neighbors(self.stat_coords, self.r2, self.distance_metric, self.weighted)
        
        if self.weighted:
            node_idx_neighbors, node_idx_distances = ball_tree_result
        else:
            node_idx_neighbors, node_idx_distances = ball_tree_result, None

        if self.verbose:
            print("done")
            
        # (3) Create network and run infomap
        if self.verbose: print("Creating network and clustering with Infomap...")
        self.stat_labels = utils.label_network(
            node_idx_neighbors, node_idx_distances, self.counts, self.weight_exponent,
            self.label_singleton, self.distance_metric, self.verbose
        )

        # (4) Reverse the downsampling in step (2)
        self.labels = self.stat_labels[inverse_indices]

        # Update model state and return labels
        self.is_fitted = True
        return self.labels


    def _hyperparam_assertions(self):
        assert self.r2 > 0, \
               "`r2` must be > 0"
        assert type(self.label_singleton) is bool, \
               "`label_singleton` either `True` or `False`"
        assert 0 <= self.min_spacial_resolution <= 1, \
               "`min_spacial_resolution` must be within [0, 1]"
        assert self.distance_metric in ['euclidean', 'haversine'], \
               "`distance_metric` should be either 'euclidean' or 'haversine'"


    def _data_assertions(self, data):
        assert not np.any(np.isnan(data)), \
               f"There are {np.isnan(np.vstack(data))} NaN values in the input data."
        assert data.shape[1] == 2, \
               "Number of columns must be 2"
        if self.distance_metric == 'haversine':
            assert (np.min(data[:, 0]) > -90 and np.max(data[:, 0]) < 90), \
                   "Latitude (column 0) must have values between -90 and 90"
            assert (np.min(data[:, 1]) > -180 and np.max(data[:, 1]) < 180), \
                   "Longitude (column 1) must have values between -180 and 180"